/* MakeTeXPK.c -- Create PK files
   Time-stamp: "97/01/03 17:57:54 mik"

   Copyright (C) 1995, 96, 97
	Christian Schenk  <cschenk@berlin.snafu.de>

   This file is part of MiKTeX.

   MiKTeX is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   MiKTeX is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with MiKTeX; if not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <io.h>
#include <direct.h>
#include <process.h>

#include "miktex.h"
#include "MakeTeXPK.rc"

#if defined (_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif /* _WIN32 */

static int verbose_p;
static int debug_p;
static int quiet_p;
static int print_only_p;
static const char *texdir;
static const char *destdir;
static char *name;
static int dpi;
static int bdpi;
static char *magnification;
static char *mode;

static char gfname[_MAX_PATH];
static char pkname[_MAX_PATH];
static char defdestdir[_MAX_PATH];
static char command_line[_MAX_PATH];

static char cwd[_MAX_PATH];
static char *temp_dir;

#if defined (_WIN32)
static char system_temp_dir[_MAX_PATH];
static int drive;
static int temp_drive;
#endif /* _WIN32 */

static struct option long_options[] =

{
  "debug",		no_argument,		0,	'd',
  "help",		no_argument,		0,	'h',
  "print-only",		no_argument,		0,	'n',
  "tex-dir",		required_argument,	0,	'T',
  "dest-dir",		required_argument,	0,	'D',
  "quiet",		no_argument,		0,	'q',
  "verbose",		no_argument,		0,	'v',
  "version",		no_argument,		0,	'V',
  0,			no_argument,		0,	0,
};

static void
show_version (void)

{
  printf ("This is MakeTeXPK version %s (%s %s)\n",
	  VER_FILEVERSION_STR, VER_PRODUCTNAME_STR,
	  VER_PRODUCTVERSION_STR);
}

static void
usage (void)

{
  puts ("Usage: MakeTeXPK [OPTION]... name dpi bdpi magnification [MODE]\n");

  puts ("This program makes a PK font.\n");

  puts ("NAME is the name of the font, such as `cmr10'.  DPI is");
  puts ("the resolution the font is needed at.  BDPI is the base");
  puts ("resolution, useful for figuring out the mode to make the font");
  puts ("in.  MAGNIFICATION is a string to pass to MF as the");
  puts ("magnification.  MODE, if supplied, is the mode to use.\n");

  puts ("Options:");
  puts ("  -d, --debug  Print debugging information.");
  puts ("  -D DIRECTORY  --dest-dir DIRECTORY  Specify the destination directory.");
  puts ("  -h  --help  Print this help screen and exit.");
  puts ("  -T DIRECTORT  --tex-dir DIRECTORY  Specify the MiKTeX root directory.");
  puts ("  -v  --verbose  Print information on what is being done.");
  puts ("  -V  --version  Print the version number and exit.");
  puts ("  -n  --print-only  Print what commands would be executed.");
}

static void
clean ()

{
#if defined (_WIN32)
  if (print_only_p)
    printf ("cd %s\n", cwd);
  else
    {
      WIN32_FIND_DATA fdata;
      HANDLE hnd = FindFirstFile ("*", &fdata);
      if (hnd != INVALID_HANDLE_VALUE)
	{
	  int proceed = 1;
	  while (proceed)
	    {
	      if (! (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
		{
		  if (remove (fdata.cFileName) != 0)
		    perror (fdata.cFileName);
		}
	      proceed = FindNextFile (hnd, &fdata);
	    }
	  FindClose (hnd);
	}
      if (_chdrive (drive) != 0)
	perror (cwd);
      if (chdir (cwd) != 0)
	perror (cwd);
      if (rmdir (temp_dir) != 0)
	perror (temp_dir);
    }
#endif /* _WIN32 */
}

#define STRDUP(s) strcpy (_alloca (strlen (s) + 1), s)

int
main (int argc,
      char ** argv)

{
  int c;
  int option_index;
  char *tmpptr;
  int rc;
  char destname[_MAX_PATH];
  char *cp, *cp1;
  static char buf[2048];
  char *pargv[20];
  int pargc;
  char mode_def[50];
  char mag_def[50];

  print_only_p = 0;
  quiet_p = 0;
#if defined (DEBUG)
  verbose_p = 1;
  debug_p = 1;
#else
  verbose_p = 0;
  debug_p = 0;
#endif
  texdir = 0;
  destdir = 0;

  c4pargv[0] = argv[0];
  while ((c=getopt_long (argc, argv, "dD:hnvV", long_options, &option_index))
	 != EOF)
    {
      switch (c)
	{
	case 'h':
	  usage ();
	  exit (0);
	  break;
	case 'n':
	  print_only_p = 1;
	  break;
	case 'T':
	  texdir = optarg;
	  break;
	case 'D':
	  destdir = optarg;
	  break;
	case 'd':
	  debug_p = 1;
	  break;
	case 'v':
	  verbose_p = 1;
	  break;
	case 'q':
	  quiet_p = 1;
	  break;
	case 'V':
	  show_version ();
	  exit (0);
	  break;
	default:
	  usage ();
	  exit (1);
	  break;
	}
    }

  if (argc-optind < 4)
    {
      usage ();
      exit (1);
    }

  name = argv[optind++];
  dpi = atoi (argv[optind++]);
  bdpi = atoi (argv[optind++]);
  magnification = argv[optind++];

  mode = 0;
  if (optind < argc)
    mode = argv[optind++];

  if (optind != argc)
    {
      usage ();
      exit (1);
    }

  if (texdir == 0)
    {
      char buf[MAX_PATH];
      texdir = get_cfg_value ("MiKTeX", "Root Directory",
			      buf, sizeof (buf), 0);
      if (texdir == 0)
	texdir = "c:\\texmf";
      else
	texdir = STRDUP (texdir);
    }

  if (getcwd (cwd, sizeof (cwd)) == 0)
    {
      perror ("getcwd");
      exit (1);
    }

#if defined (_WIN32)
  drive = _getdrive ();
  GetTempPath (sizeof (system_temp_dir), system_temp_dir);
  temp_dir = _tempnam (system_temp_dir, "pk");
  if (temp_dir == 0)
    {
      perror ("_tempnam");
      exit (1);
    }
  temp_drive = tolower (temp_dir[0]) - 'a' + 1;
#endif /* _WIN32 */

  if (mode && strcmp (mode, "default") == 0)
    mode = 0;

  if (mode == 0)
    {
      char *buf;
      switch (bdpi)
	{
	case 300: mode = "imagen"; break;
	case 600:
	  buf = _alloca (strlen (magnification)+4+1);
	  mode = "imagen";
	  sprintf (buf, "2*(%s)", magnification);
	  magnification = buf;
	  break;
	case 200: mode = "FAX"; break;
	case 360: mode = "nextII"; break;
	case 400: mode = "nexthi"; break;
	case 100: mode = "nextscreen"; break;
	case 72: mode = "seventwo"; break;
	case 635: mode = "linolo"; break;
	case 1270: mode = "linohi"; break;
	case 2540: mode = "linosuper"; break;
	default:
	  fprintf (stderr, "I don't know the mode for %d\n", bdpi);
	  fprintf (stderr, "Have your system admin update MakeTeXPK.c");
	  exit (1);
	}
    }

  if (destdir == 0)
    {
      char buf[MAX_PATH];
      const char *cp = get_cfg_value ("MakeTeXPK", "PK Temp Dir",
				      buf, sizeof (buf), 0);
      char *cp1 = defdestdir;
      if (cp == 0)
	cp = ".";
      while (*cp)
	{
	  if (*cp == '%')
	    {
	      switch (cp[1])
		{
		default:
		  *cp1 = 0;
		  break;
		case '%':
		  *cp1++ = '%';
		  *cp1 = 0;
		  break;
		case 'b':
		  sprintf (cp1, "%d", bdpi);
		  break;
		case 'd':
		  sprintf (cp1, "%d", dpi);
		  break;
		case 'm':
		  strcpy (cp1, mode);
		  break;
		}
	      cp1 += strlen (cp1);
	      cp += 2;
	    }
	  else
	    *cp1++ = *cp++;
	}
      *cp1 = 0;
      destdir = defdestdir;
    }

  sprintf (gfname, "%s.%dgf", name, dpi);
  cp = get_cfg_value ("MakeTeXPK", "PK File Name", buf, sizeof (buf), 0);
  if (cp == 0)
    cp = "%f.%dpk";
  cp1 = pkname;
  while (*cp)
    {
      if (*cp == '%')
	{
	  switch (cp[1])
	    {
	    default:
	      *cp1 = 0;
	      break;
	    case '%':
	      *cp1++ = '%';
	      *cp1 = 0;
	      break;
	    case 'f':
	      sprintf (cp1, "%s", name);
	      break;
	    case 'b':
	      sprintf (cp1, "%d", bdpi);
	      break;
	    case 'd':
	      sprintf (cp1, "%d", dpi);
	      break;
	    }
	  cp1 += strlen (cp1);
	  cp += 2;
	}
      else
	*cp1++ = *cp++;
    }
  *cp1 = 0;

  sprintf (destname, "%s/%s", destdir, pkname);
  if (_access (destname, 0) == 0)
    {
      if (! quiet_p)
	printf ("%s already exists!", destname);
      exit (0);
    }

  if (print_only_p)
    printf ("mkdir %s\n", temp_dir);
  else if (mkdir (temp_dir) != 0)
    {
      perror (temp_dir);
      exit (1);
    }

#if defined (_WIN32)
  if (print_only_p)
    ;
  else if (_chdrive (temp_drive) != 0)
    {
      perror (temp_dir);
      clean ();
      exit (1);
    }
#endif /* _WIN32 */

  if (print_only_p)
    printf ("cd %s\n", temp_dir);
  else if (chdir (temp_dir) != 0)
    {
      perror (temp_dir);
      clean ();
      exit (1);
    }

  if (verbose_p)
    printf ("Creating %s...\n", pkname);

  sprintf (command_line, "%s\\miktex\\bin\\virmf.exe", texdir);

  sprintf (mode_def, "\\mode:=%s;", mode);
  sprintf (mag_def, "\\mag:=%s;", magnification);
  
  pargc = 0;
  pargv[pargc++] = command_line;
  pargv[pargc++] = mode_def;
  pargv[pargc++] = mag_def;
  pargv[pargc++] = "batchmode;";
  pargv[pargc++] = "input";
  pargv[pargc++] = name;
  pargv[pargc++] = 0;
  
  if (! quiet_p || print_only_p)
    printf ("%s \\mode:=%s; mag:=%s; batchmode; input %s",
	    command_line, mode, magnification, name);

  if (print_only_p)
    rc = 0;
  else
    {
#if defined (_WIN32)
      _flushall ();
#endif
      rc = _spawnv (_P_WAIT, command_line, pargv);
    }

  if (rc < 0)
    {
      perror (command_line);
      clean ();
      exit (1);
    }

  if (rc > 0)
    {
      fprintf (stderr, "METAFONT failed for some reason on %s!\n", name);
      if (debug_p)
	fprintf (stderr, "Intermediate files in %s are *not* removed!\n",
		 temp_dir);
      else
	clean ();
      exit (1);
    }

  if (rc == 0)
    {
      if (! print_only_p && _access (gfname, 00) < 0)
	{
	  fprintf (stderr, "METAFONT failed for some reason on %s\n", name);
	  if (debug_p)
	    fprintf (stderr, "Intermediate files in %s are *not* removed!\n",
		     temp_dir);
	  else
	    clean ();
	  exit (1);
	}

      sprintf (command_line, "%s\\miktex\\bin\\gftopk.exe", texdir);
      pargc = 0;
      pargv[pargc++] = command_line;
      pargv[pargc++] = gfname;
      pargv[pargc++] = pkname;
      pargv[pargc++] = 0;

      if (verbose_p || print_only_p)
	printf ("%s %s %s\n", command_line, gfname, pkname);
      if (print_only_p)
	rc = 0;
      else
	{
#if defined (_WIN32)
	  _flushall ();
#endif
	  rc = _spawnv (_P_WAIT, command_line, pargv);
	}

      if (rc < 0)
	{
	  perror (command_line);
	  clean ();
	  exit (1);
	}

      if (rc > 0)
	{
	  fprintf (stderr, "gftopk failed for some reason on %s!\n", name);
	  if (debug_p)
	    fprintf (stderr, "Intermediate files in %s are *not* removed!\n",
		     temp_dir);
	  else
	    clean ();
	  exit (1);
	}

      if (rc == 0)
	{
	  if (! print_only_p && _access (pkname, 00) < 0)
	    {
	      fprintf (stderr, "gftopk failed for some reason on %s\n",
		       name);
	      if (debug_p)
		fprintf (stderr,
			 "Intermediate files in %s are *not* removed!\n",
			 temp_dir);
	      else
		clean ();
	      exit (1);
	    }
	  if (print_only_p || verbose_p)
	    printf ("move %s %s\n", pkname, destname);
	  if (! print_only_p)
	    {
	      if (make_path (destdir))
		{
		  if (rename (pkname, destname) != 0)
		    {
		      perror (pkname);
		      clean ();
		      exit (1);
		    }
		}
	    }
	}
    }

  clean ();
  return (0);
}
