/* MakeTeXTFM.c -- Create TFM files
   Time-stamp: "97/01/31 11:53:47 cschenk"

   Copyright (C) 1995, 96, 97
	Christian Schenk  <cschenk@berlin.snafu.de>

   This file is part of MiKTeX.

   MiKTeX is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   MiKTeX is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with MiKTeX; if not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <io.h>
#include <direct.h>
#include <process.h>

#include "miktex.h"
#include "MakeTeXTFM.rc"

#if defined (_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif /* _WIN32 */

static int verbose_p;
static int debug_p;
static int quiet_p;
static int print_only_p;
static const char *texdir;
static const char *localdir;
static const char *destdir;
static char *name;

static char tfmname[_MAX_PATH];
static char command_line[_MAX_PATH+300];

static char cwd[_MAX_PATH];
static char *temp_dir;

#if defined (_WIN32)
static char system_temp_dir[_MAX_PATH];
static int drive;
static int temp_drive;
#endif /* _WIN32 */

static struct option long_options[] =

{
  "debug",		no_argument,		0,	'd',
  "help",		no_argument,		0,	'h',
  "print-only",		no_argument,		0,	'n',
  "local-dir",		required_argument,	0,	'L',
  "tex-dir",		required_argument,	0,	'T',
  "dest-dir",		required_argument,	0,	'D',
  "quiet",		no_argument,		0,	'q',
  "verbose",		no_argument,		0,	'v',
  "version",		no_argument,		0,	'V',
  0,			no_argument,		0,	0,
};

static void
show_version (void)

{
  printf ("This is MakeTeXTFM version %s (%s %s)\n",
	  VER_FILEVERSION_STR, VER_PRODUCTNAME_STR,
	  VER_PRODUCTVERSION_STR);
}

static void
usage (void)

{
  puts ("Usage: MakeTeXTFM [OPTION]... name");

  puts ("This program makes a TeX font metric file.\n");

  puts ("NAME is the name of the font, such as `cmr10'.\n");

  puts ("Options:");
  puts ("  -d, --debug");
  puts ("  -D DIRECTORY, --dest-dir DIRECTORY");
  puts ("  -h, --help");
  puts ("  -L DIRECTORY, --local-dir DIRECTORY");
  puts ("  -T DIRECTORT, --tex-dir DIRECTORY");
  puts ("  -v, --verbose");
  puts ("  -V, --version");
  puts ("  -n, --print-only");
}

static void
clean ()

{
#if defined (_WIN32)
  if (print_only_p)
    printf ("cd %s\n", cwd);
  else
    {
      WIN32_FIND_DATA fdata;
      HANDLE hnd = FindFirstFile ("*", &fdata);
      if (hnd != INVALID_HANDLE_VALUE)
	{
	  int proceed = 1;
	  while (proceed)
	    {
	      if (! (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
		{
		  if (remove (fdata.cFileName) != 0)
		    perror (fdata.cFileName);
		}
	      proceed = FindNextFile (hnd, &fdata);
	    }
	  FindClose (hnd);
	}
      if (_chdrive (drive) != 0)
	perror (cwd);
      if (chdir (cwd) != 0)
	perror (cwd);
      if (rmdir (temp_dir) != 0)
	perror (temp_dir);
    }
#endif /* _WIN32 */
}

#define STRDUP(s) strcpy (_alloca (strlen (s) + 1), s)

int
main (int argc,
      char ** argv)

{
  int c;
  int option_index;
  char *tmpptr;
  int rc;
  char destname[_MAX_PATH];
  char *pargv[10];
  int pargc;

  print_only_p = 0;
  quiet_p = 0;
#if defined (DEBUG)
  verbose_p = 1;
  debug_p = 1;
#else
  verbose_p = 0;
  debug_p = 0;
#endif
  texdir = 0;
  localdir = 0;
  destdir = 0;

  c4pargv[0] = argv[0];
  while ((c=getopt_long (argc, argv, "dD:hnvV", long_options, &option_index))
	 != EOF)
    {
      switch (c)
	{
	case 'h':
	  usage ();
	  exit (0);
	  break;
	case 'n':
	  print_only_p = 1;
	  break;
	case 'T':
	  texdir = optarg;
	  break;
	case 'L':
	  localdir = optarg;
	  break;
	case 'D':
	  destdir = optarg;
	  break;
	case 'd':
	  debug_p = 1;
	  break;
	case 'v':
	  verbose_p = 1;
	  break;
	case 'q':
	  quiet_p = 1;
	  break;
	case 'V':
	  show_version ();
	  exit (0);
	  break;
	default:
	  usage ();
	  exit (1);
	  break;
	}
    }

  if (argc - optind < 1)
    {
      usage ();
      exit (1);
    }

  name = argv[optind++];

  if (optind != argc)
    {
      usage ();
      exit (1);
    }

  if (texdir == 0)
    {
      char buf[MAX_PATH];
      texdir = get_cfg_value ("MiKTeX", "Root Directory", buf, sizeof (buf),
			      0);
      if (texdir == 0)
	texdir = "c:\\texmf";
      else
	texdir = STRDUP (texdir);
    }

  if (localdir == 0)
    localdir = texdir;

  if (getcwd (cwd, sizeof (cwd)) == 0)
    {
      perror ("getcwd");
      exit (1);
    }

#if defined (_WIN32)
  drive = _getdrive ();
  GetTempPath (sizeof (system_temp_dir), system_temp_dir);
  temp_dir = _tempnam (system_temp_dir, "tfm");
  if (temp_dir == 0)
    {
      perror ("_tempnam");
      exit (1);
    }
  temp_drive = tolower (temp_dir[0]) - 'a' + 1;
#endif /* _WIN32 */

  if (destdir == 0)
    {
      char buf[MAX_PATH];
      destdir = get_cfg_value ("TeX", "Font Metric Temp Dir", buf,
			       sizeof (buf), 0);
      if (destdir == 0)
	destdir = ".";
      else
	destdir = STRDUP (destdir);
    }

  sprintf (tfmname, "%s.tfm", name);
  sprintf (destname, "%s/%s", destdir, tfmname);
  if (_access (destname, 0) == 0)
    {
      if (! quiet_p)
	printf ("%s already exists!", destname);
      exit (0);
    }

  if (print_only_p)
    printf ("mkdir %s\n", temp_dir);
  else if (mkdir (temp_dir) != 0)
    {
      perror (temp_dir);
      exit (1);
    }

#if defined (_WIN32)
  if (print_only_p)
    ;
  else if (_chdrive (temp_drive) != 0)
    {
      perror (temp_dir);
      clean ();
      exit (1);
    }
#endif /* _WIN32 */

  if (print_only_p)
    printf ("cd %s\n", temp_dir);
  else if (chdir (temp_dir) != 0)
    {
      perror (temp_dir);
      clean ();
      exit (1);
    }

  if (verbose_p)
    printf ("Creating %s...\n", name);

  sprintf (command_line, "%s\\miktex\\bin\\virmf.exe", texdir);

  pargc = 0;
  pargv[pargc++] = command_line;
  pargv[pargc++] = "\\mode:=cx;";
  pargv[pargc++] = "batchmode;";
  pargv[pargc++] = "input";
  pargv[pargc++] = name;
  pargv[pargc++] = 0;

  if (! quiet_p || print_only_p)
    printf ("%s \\mode:=cx; batchmode; input %s\n", command_line, name);

  if (print_only_p)
    rc = 0;
  else
    {
#if defined (_WIN32)
      _flushall ();
#endif
      rc = _spawnv (_P_WAIT, command_line, pargv);
    }

  if (rc < 0)
    {
      perror (command_line);
      clean ();
      exit (1);
    }

  if (rc > 0)
    {
      fprintf (stderr, "METAFONT failed for some reason on %s!\n", name);
      if (debug_p)
	fprintf (stderr, "Intermediate files in %s are *not* removed!\n",
		 temp_dir);
      else
	clean ();
      exit (1);
    }
  else
    {
      if (! print_only_p && _access (tfmname, 00) < 0)
	{
	  fprintf (stderr, "METAFONT failed for some reason on %s\n", name);
	  if (debug_p)
	    fprintf (stderr, "Intermediate files in %s are *not* removed!\n",
		     temp_dir);
	  else
	    clean ();
	  exit (1);
	}
      if (print_only_p || verbose_p)
	printf ("move %s %s\n", tfmname, destname);
      if (! print_only_p)
	{
	  if (make_path (destdir))
	    {
	      if (rename (tfmname, destname) != 0)
		{
		  perror (tfmname);
		  clean ();
		  exit (1);
		}
	    }
	}
    }

  clean ();
  return (0);
}
