/* configure.c -- MiKTeX configuration utility
   Time-stamp: "97/01/31 17:02:42 cschenk"

   Copyright (C) 1995, 1996, 1997
		Christian Schenk <cschenk@berlin.snafu.de>

   This file is part of MiKTeX.

   MiKTeX is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   MiKTeX is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with MiKTeX; if not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include <stdio.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <getopt.h>

#include "miktex.h"
#include "configure.rc"

#if defined (_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif /* _WIN32 */

#define E_UNKNOWN_ROOT "the TeX root directory is still undefined"

static int verbose_p;
static int debug_p;
static int quiet_p;
static int print_only_p;
static char *root_dir;

static struct option long_options[] =

{
  "debug",		no_argument,		0,	'd',
  "help",		no_argument,		0,	'h',
  "list-modes",		no_argument,		0,	1,
  "update-fndb",	no_argument,		0,	'u',
  "print-only",		no_argument,		0,	'n',
  "root-directory",	required_argument,	0,	'r',
  "quiet",		no_argument,		0,	'q',
  "config",		required_argument,	0,	'c',
  "verbose",		no_argument,		0,	'v',
  "version",		no_argument,		0,	'V',
  0,			no_argument,		0,	0,
};

static void
show_version (void)

{
  printf ("This is configure version %s (%s %s)\n",
	  VER_FILEVERSION_STR, VER_PRODUCTNAME_STR,
	  VER_PRODUCTVERSION_STR);
}

static void
usage (void)

{
  puts ("Usage: configure [OPTION]...\n");
  puts ("This program configures MiKTeX for your system.\n");
  puts ("Options:");
  puts ("  -c FILENAME  --config FILENAME  Specify the config file.");
#if 0
  puts ("  -d  --debug  Print debugging information.");
#endif
  puts ("  -u  --update-fndb  Refresh the filename database.");
  puts ("  -h  --help  Print this help screen and exit.");
  puts ("  -r DIR  --root-directory DIR  Specify the TeX root directory.");
  puts ("  -v  --verbose  Print information on what is being done.");
  puts ("  -V  --version  Print the version number and exit.");
  puts ("  -n  --print-only  Print what would be done.");
}

static void
my_fatal_error (const char *msg)

{
  fprintf (stderr, "configure: %s\n", msg);
  exit (1);
}

static void
fatal_file_error (const char *filename)

{
  fprintf (stderr, "configure: ");
  perror (filename);
  exit (1);
}

static void
parse_line (char *a_line,
	    char **key,
	    char **value,
	    char **comment)

{
  if (key)
    *key = 0;
  if (value)
    *value = 0;
  if (isalnum (*a_line))
    {
      if (key)
	*key = a_line;
      if ((a_line = strchr (a_line, '=')) != 0)
	{
	  char *cmnt;

	  if (value)
	    *value = a_line + 1;
	  if ((cmnt = strchr (a_line, ';')) != 0 &&
	      isspace (cmnt[-1]))
	    {
	      *cmnt = 0;
	      if (comment)
		*comment = cmnt+1;
	      cmnt++;
	      cmnt += strlen (cmnt) - 1;
	      if (*cmnt == '\n')
		*cmnt = 0;
	    }
	  do
	    {
	      a_line--;
	    }
	  while ((*a_line == ' ') || (*a_line == '\t'));
	  *++a_line = 0;
	  if (value)
	    {
	      while ((*(*value) == ' ') || (*(*value) == '\t'))
		(*value)++;
	      a_line = &((*value)[ strlen (*value) - 1 ]);
	      while ((*a_line == ' ') || (*a_line == '\t') ||
		     (*a_line == '\n'))
		a_line--;
	      *++a_line = 0;
	    }
	}
    }
}

static void
expand (const char *value,
	char *xvalue)

{
  size_t xlen = strlen (root_dir);
  while (*value)
    {
      if (value[0] == '%')
	{
	  switch (value[1])
	    {
	    default:
	      *xvalue++ = '?';
	      break;
	    case '%':
	      *xvalue++ = '%';
	      break;
	    case 'R':
	      memcpy (xvalue, root_dir, xlen);
	      xvalue += xlen;
	    }
	  value += 2;
	}
      else
	*xvalue++ = *value++;
    }
  *xvalue = 0;
}

static int
fnamecmp (const char **s1,
	  const char **s2)

{
  return (_stricmp (*s1, *s2));
}

static void
dump_listing (FILE *fndb_stream,
	      const char *parent_root,
	      const char *dirname)

{
  WIN32_FIND_DATA data;
  HANDLE hnd;
  char *search_pattern;
  char *dir_path;
  if (verbose_p)
    printf ("  %s...\n", dirname);
  search_pattern = (char *) _alloca (strlen (parent_root)
				     + strlen (dirname) + 4);
  sprintf (search_pattern, "%s/%s/*", parent_root, dirname);
  hnd = FindFirstFile (search_pattern, &data);
  dir_path = search_pattern;
  dir_path[strlen (dir_path) - 2] = 0;
  if (hnd != INVALID_HANDLE_VALUE)
    {
      size_t nsubdirs = 0;
      size_t nfiles = 0;
      static char **subdirs = 0;
      static char **files = 0;
      static size_t max_subdirs = 0;
      static size_t max_files = 0;
      static size_t subdir_base = 0;
      size_t i;
      BOOL go_on;
      size_t this_subdir_base = subdir_base;
      do
	{
	  if (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	    {
	      if (strcmp (data.cFileName, ".") != 0
		  && strcmp (data.cFileName, "..") != 0)
		{
		  size_t l;
		  if (subdirs == 0)
 		    {
#ifdef DEBUG
		      max_subdirs = 10;
#else
		      max_subdirs = 50;
#endif
		      subdirs = (char **) malloc (max_subdirs
						  * sizeof (char *));
		    }
		  if (this_subdir_base + nsubdirs == max_subdirs)
		    {
#ifdef DEBUG
		      max_subdirs += 10;
#else
		      max_subdirs += 50;
#endif
		      subdirs = (char **) realloc (subdirs,
						   max_subdirs
						   * sizeof (char *));
		    }
		  subdirs[this_subdir_base + nsubdirs++]
		    = strdup (data.cFileName);
		}
	    }
	  else
	    {
	      if (files == 0)
		{
#if defined (DEBUG)
		  max_files = 10;
#else
		  max_files = 200;
#endif
		  files = (char **) malloc (max_files
					    * sizeof (char *));
		}
	      if (nfiles == max_files)
		{
		  max_files *= 2;
		  files = (char **) realloc (files,
					     max_files
					     * sizeof (char *));
		}
	      files[nfiles++] = strdup (data.cFileName);
	    }
	  go_on = FindNextFile (hnd, &data);
	}
      while (go_on);
      FindClose (hnd);
      fprintf (fndb_stream,
	       "%d %d %s\n", (int) nsubdirs, (int) nfiles,
	       (strcmp (dirname, ".") == 0 ? parent_root : dirname));
      qsort (files, nfiles, sizeof (files[0]), fnamecmp);
      for (i = 0; i < nfiles; i++)
	{
	  fprintf (fndb_stream, "%s\n", files[i]);
	  free (files[i]);
	}
      qsort (&subdirs[subdir_base], nsubdirs, sizeof (subdirs[0]), fnamecmp);
      subdir_base += nsubdirs;
      for (i = this_subdir_base; i < this_subdir_base + nsubdirs; i++)
	{
	  dump_listing (fndb_stream, dir_path, subdirs[i]);
	  free (subdirs[i]);
	}
      subdir_base = this_subdir_base;
    }
}

static void
make_fndb ()

{
  FILE *fndb_stream;
  char *fndb_filename = (char *) malloc (strlen (root_dir)
					 + strlen (REL_FNDB_FILENAME)
					 + 2);
  sprintf (fndb_filename, "%s/%s", root_dir, REL_FNDB_FILENAME);
  fndb_stream = fopen (fndb_filename, "wb");
  if (fndb_stream == 0)
    fatal_file_error (fndb_filename);
  else
    {
      if (verbose_p)
	printf ("now creating file name database\n");
      dump_listing (fndb_stream, root_dir, ".");
      fclose (fndb_stream);
      set_cfg_value ("fndb", "updateflag", "false");
    }
}

static void
list_modes ()

{
  mf_mode modes[1000];
  int n = get_metafont_modes (modes, 1000);
  int i;
  for (i = 0; i < n; i++)
    printf ("%-8s  %5dx%-5d  %s\n",
	    modes[i].mnemonic,
	    modes[i].hor_res,
	    modes[i].vert_res,
	    modes[i].description);
}

static void
remove_105_keys ()

{
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\BibTeX");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\dvips");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\fndb");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\LaTeX");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\MakeIndex");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\MakeTeXPK");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\METAFONT");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\MetaPost");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\MiKTeX");
  RegDeleteKey (HKEY_LOCAL_MACHINE, "SOFTWARE\\MiK\\MiKTeX\\TeX");
}

int main (int argc,
	  char **argv)

{
  int ret = 0;
  char a_line[2000];
  char section[200];
  char envfilename[260];
  char *key;
  char *value;
  FILE *envfile;
  char xvalue[2000];
  int c;
  int option_index;
  int make_fndb_flag;
  char path[200];
  int alt_config;

  alt_config = 0;
  print_only_p = 0;
  verbose_p = 0;
  quiet_p = 0;
  debug_p = 0;
  root_dir = 0;
  make_fndb_flag = 0;

#if defined (_WIN32)
  strcpy (path, "Software\\MiK\\MiKTeX\\CurrentVersion\\TeX");
#endif

  remove_105_keys ();

  option_index = 0;
  while ((c=getopt_long (argc, argv, "c:dhnr:uvV",
			 long_options, &option_index))
	 != EOF)
    {
      switch (c)
	{
	case 1:
	  list_modes ();
	  exit (0);
	  break;
	case 'c':
	  alt_config = 1;
	  strcpy (envfilename, optarg);
	  break;
	case 'h':
	  usage ();
	  exit (0);
	  break;
	case 'n':
	  print_only_p = 1;
	  break;
	case 'd':
	  debug_p = 1;
	  break;
	case 'r':
	  root_dir = strdup (optarg);
	  break;
	case 'v':
	  verbose_p = 1;
	  break;
	case 'q':
	  quiet_p = 1;
	  break;
	case 'V':
	  show_version ();
	  exit (0);
	  break;
	case 'u':
	  make_fndb_flag++;
	  break;
	default:
	  usage ();
	  exit (1);
	  break;
	}
    }

  if (argc - optind > 0)
    {
      usage ();
      exit (1);
    }

  if (root_dir == 0)
    {
#if defined (_WIN32)
      HKEY hKey;
      LONG res;
      char path[_MAX_PATH];
      sprintf (path, "Software\\MiK\\MiKTeX\\CurrentVersion\\MiKTeX");
      res = RegOpenKeyEx (HKEY_LOCAL_MACHINE,
			  path,
			  0,
			  KEY_READ,
			  &hKey);
      if (res != ERROR_SUCCESS)
	my_fatal_error (E_UNKNOWN_ROOT);
      else
	{
	  DWORD type;
	  static BYTE value[2048];
	  DWORD len = sizeof(value);
	  res = RegQueryValueEx(hKey,
				"Root Directory",
				0,
				&type,
				value,
				&len);
	  RegCloseKey (hKey);
	  if (res != ERROR_SUCCESS)
	    my_fatal_error (E_UNKNOWN_ROOT);

	  root_dir = strdup (value);
	}
#else
      my_fatal_error (E_UNKNOWN_ROOT);
#endif
    }

  if (verbose_p)
    printf ("opening MiKTeX configuration file...\n");

  if (! alt_config)
    sprintf (envfilename, "%s/miktex/config/miktex.environment", root_dir);
  envfile = fopen (envfilename, "r");
  if (envfile == 0)
    fatal_file_error (envfilename);

  while (fgets (a_line, sizeof (a_line)-1, envfile))
    {
      if (*a_line == '[')
	{
	  const char *section = strtok (a_line+1, "]");
	  if (section)
	    {
	      if (verbose_p)
		printf ("reading section [%s]...", section);
#if defined (_WIN32)
	      sprintf (path, "Software\\MiK\\MiKTeX\\CurrentVersion\\%s",
		       section);
#endif
	    }
	}
      else
	{
	  parse_line (a_line, &key, &value, 0);
	  if (key && value)
	    {
#if defined (_WIN32)
	      LONG res;
	      HKEY hKey;
	      DWORD disp;
	      DWORD len;
#endif /* _WIN32 */

	      expand (value, xvalue);

#if defined (_WIN32)
	      len = strlen (xvalue)+1;
	      if (print_only_p)
		printf ("HKEY_LOCAL_MACHINE\\%s\\%s=%s\n",
			path, key, xvalue);
	      if (! print_only_p)
		{
		  res = RegCreateKeyEx (HKEY_LOCAL_MACHINE,
					path,
					0,
					"",
					REG_OPTION_NON_VOLATILE,
					KEY_ALL_ACCESS,
					0,
					&hKey,
					&disp);
		  if (res != ERROR_SUCCESS)
		    {
		      fprintf (stderr, "Cannot create key '%s'!\n", path);
		      exit (2);
		    }
		}
	      if (! print_only_p)
		{
		  res = RegSetValueEx (hKey,
				       key,
				       0,
				       REG_SZ,
				       xvalue,
				       len);
		  if (res != ERROR_SUCCESS)
		    {
		      fprintf (stderr, "Cannot set value '%s'!\n", xvalue);
		      RegCloseKey (hKey);
		      exit (3);
		    }
		  RegCloseKey (hKey);
		}
#endif /* !_WIN32 */
	    }
	}
    }

  fclose (envfile);

  if (make_fndb_flag)
    make_fndb ();

  if (verbose_p)
    printf ("MiKTeX has been successfully configured for your system!");

  return (ret);
}
