/* miktex.c - MiKTeX Support Routines.
   Time-stamp: "97/02/14 11:54:14 cschenk"

   Copyright (C) 1991, 92, 93, 94, 95, 96, 97
	Christian Schenk  <cschenk@berlin.snafu.de>

   This file is part of MiKTeX.

   MiKTeX is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   MiKTeX is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with MiKTeX; if not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#if defined (_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <process.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "c4plib.h"
#include "miktex.h"

#if defined (_WIN32)
#if defined (IN_MIKTEX_DLL)
#define _MIKTEX_1 __declspec (dllexport)
#endif /* IN_MIKTEX_DLL */
#define _MIKTEX_2 __stdcall
#endif /* _WIN32 */

#if ! defined (_MIKTEX_1)
#define _MIKTEX_1
#endif /* ! _MIKTEX_1 */

#if ! defined (_MIKTEX_2)
#define _MIKTEX_2
#endif /* ! _MIKTEX_2 */

#include "fnamedb.h"
#include "miktex.rc"

#define PATH_SEP_CHAR_AS_STRING ";"
#define DIRNAME_DELIMITER '/'
#define RECURSION_INDICATOR "//"

#define isslash(c) ((c) == '/' || (c) == '\\')
#define ispath(filename) (strchr (filename, '/') || strchr (filename, '\\'))

#define S_ISDIR(m) (((m) & _S_IFMT) == _S_IFDIR)

static clock_t start_clock;

#ifdef DEBUG
int debug_flag = 1;
#else
int debug_flag = 0;
#endif

static void
system_error ()

{
  LPVOID lpMsgBuf;
  FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		 NULL,
		 GetLastError (),
		 MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
		 (LPTSTR) &lpMsgBuf,
		 0,
		 NULL);
  fprintf (stderr, "\n%s\n", lpMsgBuf);
  LocalFree (lpMsgBuf);
}

_MIKTEX_1 void _MIKTEX_2
fatal_error (const char *fmt,
	     ...)

{
  va_list marker;
  va_start (marker, fmt);
  fprintf (stderr, "%s: ", c4pargv[0]);
  vfprintf (stderr, fmt, marker);
  va_end (marker);
  fputc ('\n', stderr);
  exit (EXIT_FAILURE);
}

_MIKTEX_1 int _MIKTEX_2
set_cfg_value (const char *section,
	       const char *key,
	       const char *value)

{
#if defined (_WIN32)
  HKEY hKey;
  LONG res;
  char path[_MAX_PATH];
  DWORD disp;
  DWORD len;
  sprintf (path, "Software\\MiK\\MiKTeX\\CurrentVersion\\%s", section);
  res = RegCreateKeyEx (HKEY_LOCAL_MACHINE,
			path,
			0,
			"",
			REG_OPTION_NON_VOLATILE,
			KEY_ALL_ACCESS,
			0,
			&hKey,
			&disp);
  if (res != ERROR_SUCCESS)
    {
      system_error ();
      return (-1);
    }
  len = strlen (value) + 1;
  res = RegSetValueEx (hKey,
		       key,
		       0,
		       REG_SZ,
		       value,
		       len);
  if (res != ERROR_SUCCESS)
    {
      system_error ();
      RegCloseKey (hKey);
      return (-1);
    }
  RegCloseKey (hKey);
#endif
  return (0);
}

_MIKTEX_1 char * _MIKTEX_2
get_cfg_value (const char *section,
	       const char *key,
	       char *buf,
	       size_t bufsize,
	       char *defval)

{
#if defined (_WIN32)
  HKEY hKey;
  LONG res;
  char path[_MAX_PATH];
  sprintf (path, "Software\\MiK\\MiKTeX\\CurrentVersion\\%s", section);
  res = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		     path,
		     0,
		     KEY_READ,
		     &hKey);
  if (res == ERROR_SUCCESS)
    {
      DWORD type;
      DWORD len = bufsize;
      res = RegQueryValueEx(hKey,
			    key,
			    0,
			    &type,
			    buf,
			    &len);
      RegCloseKey (hKey);
      if (res == ERROR_SUCCESS)
	return (buf);
      else
	return (defval);
    }
  else
    return (defval);
#endif /* _WIN32 */
}

_MIKTEX_1 void _MIKTEX_2
mtfirstword (void)

{
#ifndef MIKTEX_DLL
  char buf[10];
  const char *cp = get_cfg_value ("miktex", "debug", buf, sizeof (buf), 0);
  if (cp && _stricmp (cp, "true") == 0)
    debug_flag = 1;
  else
    debug_flag = 0;
#endif
  if (debug_flag)
    start_clock = clock ();
}

_MIKTEX_1 void _MIKTEX_2
mtafterinit (void)

{
  ;
}

_MIKTEX_1 void _MIKTEX_2
mtlastword (void)

{
  if (debug_flag)
    {
      clock_t ellapsed_time = (clock () - start_clock) / CLOCKS_PER_SEC;
      printf ("\n%u'%02u\n", (unsigned) (ellapsed_time / 60),
	      (unsigned) (ellapsed_time % 60));
    }
}

static int
ext_find_file (const char *curdir,
	       const char *subdir,
	       const char *searchspec,
	       char *result)

{
  size_t curdirlen = strlen (curdir);
  if (subdir)
    {
#if defined (_WIN32)
      WIN32_FIND_DATA find_file_data;
      HANDLE hnd;
#endif
      size_t bufsize;
      char *buf;
      char *newcurdir;
      bufsize = curdirlen + strlen (subdir) + 10;
      buf = _alloca (bufsize);
      newcurdir= buf;
      strcpy (newcurdir, curdir);
      newcurdir[curdirlen] = '/';
      strcpy (newcurdir + curdirlen + 1, subdir);
      if (access (newcurdir, 0) == 0)
	{
	  if (ext_find_file (newcurdir, 0, searchspec, result))
	    return (1);
	}
#if defined (_WIN32)      
      strcpy (&newcurdir[curdirlen + 1], "*");
      hnd = FindFirstFile (newcurdir, &find_file_data);
      if (hnd != INVALID_HANDLE_VALUE)
	{
	  int proceed = 1;
	  while (proceed)
	    {
	      if (find_file_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY
		  && strcmp (find_file_data.cFileName, ".") != 0
		  && strcmp (find_file_data.cFileName, "..") != 0)
		{
		  strcpy (newcurdir + curdirlen + 1, find_file_data.cFileName);
		  if (ext_find_file (newcurdir, subdir, searchspec, result))
		    {
		      FindClose (hnd);
		      return (1);
		    }
		}
	      proceed = FindNextFile (hnd, &find_file_data);
	    }
	  FindClose (hnd);
	}
#endif
      return (0);
    }
  else
    {
      char *slsl = strstr (searchspec, "//");
      if (slsl)
	{
	  size_t len = slsl - searchspec;
	  char *subdir = _alloca (len + 1);
	  strncpy (subdir, searchspec, len);
	  subdir[len] = 0;
	  if (curdir && *curdir)
	    return (ext_find_file (curdir, subdir, slsl + 2, result));
	  else
	    return (ext_find_file (subdir, 0, slsl + 2, result));
	}
      else
	{
	  char *filename = result;
	  strcpy (filename, curdir);
	  strcat (filename, "/");
	  strcat (filename, searchspec);
	  if (access (filename, 0) == 0)
	    return (1);
	  else
	    {
#if defined (_WIN32)
	      WIN32_FIND_DATA find_file_data;
	      HANDLE hnd;
#endif
	      size_t bufsize = strlen (curdir) + 100; /* FIXME */
	      char *buf = _alloca (bufsize);
	      strcpy (buf, curdir);
#if defined (_WIN32)
	      strcat (buf, "/*");
	      hnd = FindFirstFile (buf, &find_file_data);
	      if (hnd != INVALID_HANDLE_VALUE)
		{
		  int proceed = 1;
		  while (proceed)
		    {
		      if (find_file_data.dwFileAttributes
			  & FILE_ATTRIBUTE_DIRECTORY
			  && strcmp (find_file_data.cFileName, ".") != 0
			  && strcmp (find_file_data.cFileName, "..") != 0)
			{
			  strcpy (buf + curdirlen + 1,
				  find_file_data.cFileName);
			  if (ext_find_file (buf, 0, searchspec, result))
			    {
			      FindClose (hnd);
			      return (1);
			    }
			}
		      proceed = FindNextFile (hnd, &find_file_data);
		    }
		}
	      FindClose (hnd);
	      return (0);
#endif
	    }
	}
    }
}

_MIKTEX_1 FILE * _MIKTEX_2
open_file (const char *file_name,
	   const char *mode)

{
  int fd;
  FILE *fp;
  int flags = (strchr (mode, 'r')
	       ? _O_RDONLY | _O_SEQUENTIAL
	       : _O_WRONLY | _O_CREAT | _O_TRUNC);
  if (strchr (mode, 'b'))
    flags |= _O_BINARY;
  else
    flags |= _O_TEXT;
  fd = _open (file_name, flags, _S_IREAD | _S_IWRITE);
  if (fd >= 0)
    fp = _fdopen (fd, mode);
  else
    fp = 0;
  if (fp)
    setvbuf (fp, 0, _IOFBF, 1024 * 4);
  return (fp);
}

_MIKTEX_1 char * _MIKTEX_2
get_home_directory (char *buf)

{
  const char *homedrive = getenv ("HOMEPATH");
  const char *homepath = getenv ("HOMEPATH");
  const char *home = getenv ("HOME");
  static char mybuf[_MAX_PATH];
  if (buf == 0)
    buf = mybuf;
  if (homedrive && homepath)
    sprintf (buf, "%s\\%s", homedrive, homepath);
  else if (home)
    strcpy (buf, home);
  else
    strcpy (buf, "c:\\");
  return (buf);
}

static boolean
external_find_file (const char *filename,
		    const char *path_list,
		    char *result)

{
  char *my_path_list = strdup (path_list);
  char *path;
  char *next_path;
  int found = 0;
  path = my_path_list;
  next_path = strchr (path, ';');
  if (next_path)
    *next_path++ = 0;
  while (path && ! found)
    {
      size_t pathlen;
      size_t bufsize;
      char *buf;
      char *slsl;
      if (path[0] == '~')
	path = get_home_directory (0);
      pathlen = strlen (path);
      bufsize = pathlen + strlen (filename) + 2;
      buf = _alloca (bufsize);
      strcpy (buf, path);
      if (buf[pathlen - 1] != '/' && buf[pathlen - 1] != '\\')
	strcat (buf, "/");
      strcat (buf, filename);
      slsl = strstr (buf, "//");
      if (slsl == 0)
	{
	  strcpy (result, buf);
	  if (access (result, 0) == 0)
	    {
	      found = 1;
	    }
	}
      else
	{
	  char *curdir;
	  char *searchspec;
	  curdir = buf;
	  *slsl = 0;
	  searchspec = slsl + 2;
	  if (ext_find_file (curdir, 0, searchspec, result))
	    found = 1;
	}
      if (next_path && *next_path)
	{
	  path = next_path;
	  next_path = strchr (next_path, ';');
	  if (next_path)
	    *next_path++ = 0;
	}
      else
	path = 0;
    }
  free (my_path_list);
  return (found);
}

static int
int_find_file (int dir_handle,
	       const char *subdir,
	       const char *searchspec,
	       char *result)

{
  int subdir_handle;
  if (subdir)
    {
      if ((subdir_handle = fndb_open_directory (dir_handle, subdir)) >= 0)
	{
	  int found = int_find_file (subdir_handle, 0, searchspec, result);
	  fndb_close_directory (subdir_handle);
	  if (found)
	    return (1);
	}
      while ((subdir_handle = fndb_open_next_sub_directory (dir_handle)) >= 0)
	{
	  int found = int_find_file (subdir_handle, subdir, searchspec,
				     result);
	  fndb_close_directory (subdir_handle);
	  if (found)
	    return (1);
	}
    }
  else
    {
      char *slsl = strstr (searchspec, "//");
      if (slsl)
	{
	  size_t len = slsl - searchspec;
	  char *subdir = _alloca (len + 1);
	  strncpy (subdir, searchspec, len);
	  subdir[len] = 0;
	  return (int_find_file (dir_handle, subdir, slsl + 2, result));
	}
      else
	{
	  if (fndb_find_file (dir_handle, searchspec, result, 0))
	    return (1);
	  else
	    {
	      while ((subdir_handle
		      = fndb_open_next_sub_directory (dir_handle)) >= 0)
		{
		  int found = int_find_file (subdir_handle, 0,
					     searchspec, result);
		  fndb_close_directory (subdir_handle);
		  if (found)
		    return (1);
		}
	    }
	}
    }
  return (0);
}

static fndb_stat = 0;

static int
open_fndb ()

{
  char fndb_filename[300];
  int ret = -1;
#if 0
  int updateflag = 0;
  int noautoupdate = 1;
#endif
  char *cp;
  char root[300];
  char buf[50];
#if 0
  cp = get_cfg_value ("fndb", "updateflag", buf, sizeof (buf), "false");
  if (_stricmp (cp, "true") == 0)
    updateflag = 1;
  cp = get_cfg_value ("fndb", "noautoupdate", buf, sizeof (buf), "true");
  if (_stricmp (cp, "false") == 0)
    noautoupdate = 0;
#endif
  if (get_cfg_value ("MiKTeX", "Root Directory", root, sizeof (root), 0) == 0)
    {
      if (debug_flag)
	fprintf (stderr,
		 "cannot read texmf.fndb: unknown root directory\n");
    }
  else
    {
      strcpy (fndb_filename, root);
      strcat (fndb_filename, "/");
      strcat (fndb_filename, REL_FNDB_FILENAME);
#if 0
      if (updateflag && ! noautoupdate)
	{
	  char command_line[_MAX_PATH];
	  char *argv[10];
	  int argc = 0;
	  strcpy (command_line, root);
	  strcat (command_line, "\\miktex\\config\\configure.exe");
	  argv[argc++] = command_line;
	  argv[argc++] = "--update";
	  if (debug_flag)
	    argv[argc++] = "--verbose";
	  argv[argc++] = 0;
	  if (debug_flag)
	    {
	      fprintf (stderr, "updating filename database...\n");
	      puts (command_line);
	    }
	  _spawnv (_P_WAIT, command_line, argv);
#if 0
	  set_cfg_value ("fndb", "updateflag", "false"); /* fixme */
#endif
	  updateflag = 0;
	}
#endif
      if (access (fndb_filename, 0) != 0)
	fprintf (stderr, "note: texmf.fndb does not exist\n");
      else
	{
#if 0
	  if (updateflag)
	    fprintf (stderr, "note: texmf.fndb is out of date\n");
#endif
	  ret = (fndb_new (fndb_filename, root) == 0 ? 1 : -1);
	}
    }
  return (ret);
}

static int
internal_find_file (const char *filename,
		    const char *path_list,
		    char *result)

{
  size_t filename_len = strlen (filename);
  char *my_path_list = strdup (path_list);
  char *path;
  char *next_path;
  int found = 0;
  if (fndb_stat == 0)
    fndb_stat = open_fndb ();
  if (fndb_stat < 0)
    return (0);
  path = my_path_list;
  next_path = strchr (path, ';');
  if (next_path)
    *next_path++ = 0;
  while (path && ! found)
    {
      size_t path_len = strlen (path);
      char *slsl;
      char *search_spec = _alloca (path_len + filename_len + 3);
      strcpy (search_spec, path);
      if (search_spec[path_len - 1] != '/'
	  && search_spec[path_len - 1] != '\\')
	strcat (search_spec, "/");
      strcat (search_spec, filename);

      slsl = strstr (search_spec, "//");
      if (slsl == 0)
	{
	  int hnd;
	  strcpy (result, search_spec);
	  if (fndb_find_file (0, result, 0, 0))
	    {
	      found = 1;
	    }
	}
      else
	{
	  if (int_find_file (0, 0, search_spec, result))
	    found = 1;
	}
      if (next_path && *next_path)
	{
	  path = next_path;
	  next_path = strchr (next_path, ';');
	  if (next_path)
	    *next_path++ = 0;
	}
      else
	path = 0;
    }
  free (my_path_list);
  return (found);
}

/* "Find file" machinery. */
_MIKTEX_1 boolean _MIKTEX_2
find_file (const char *filename,
	   const char *path_list,
	   char *result)
     
{
  boolean found;
  if (ispath (filename)
      || (path_list && path_list[0] == '.' && path_list[1] == ';'))
    {
      if (_access (filename, 0) == 0)
	{
	  strcpy (result, filename);
	  return (1);
	}
      else if (ispath (filename))
	return (0);
    }
  found = internal_find_file (filename, path_list, result);
  if (found && _access (result, 0) != 0)
    {
      found = 0;
#if 0
      set_cfg_value ("fndb", "updateflag", "true");
#endif
    }
  if (! found)
    {
      if (debug_flag)
	fprintf (stderr, "%s not found in fndb\n", filename);
      found = external_find_file (filename, path_list, result);
      if (found)
	{
	  if (debug_flag)
	    fprintf (stderr, "found %s\n", result);
#if 0
	  if (! (*result == '.'
		 && ((result[1] == '.' && isslash (result[2]))
		     || isslash (result[1]))))
	    set_cfg_value ("fndb", "updateflag", "true");
#endif
	}
    }
  return (found);
}

typedef struct
{
  FILE *_c4p_fp;
  unsigned char _c4p_buf;
} bytefile;

_MIKTEX_1 boolean _MIKTEX_2
make_path (const char *path)

/* borrowed from GNU File Utilities! */

{
  char *dirpath;
  struct stat stats;
  int retval = 0;

  dirpath = (char *) _alloca (strlen (path) + 1);
  strcpy (dirpath, path);

  if (stat (dirpath, &stats))
    {
      char *slash;
      char *cp;
      struct ptr_list
      {
	char *dirname_end;
	struct ptr_list *next;
      };
      struct ptr_list *p, *leading_dirs = NULL;

      slash = dirpath;
      if (slash[0] && slash[1] == ':')
	slash += 2;
      while (*slash == '/' || *slash == '\\')
	slash++;
      while ((cp = strchr (slash, '/')) || (cp = strchr (slash, '\\')))
	{
	  slash = cp;
	  *slash = '\0';
	  if (stat (dirpath, &stats))
	    {
	      if (mkdir (dirpath))
		{
		  fatal_error ("cannot make directory '%s'", dirpath);
		  return (false);
		}
	    }
	  else if (!S_ISDIR (stats.st_mode))
	    {
	      fatal_error ("'%s' exists but is not a directory", dirpath);
	      return (true);
	    }

	  *slash++ = '/';

	  while (*slash == '/' || *slash == '\\')
	    slash++;
	}
      if (mkdir (dirpath))
	{
	  fatal_error ("cannot make directory '%s'", dirpath);
	  return (false);
	}
    }
  else
    {
      if (!S_ISDIR (stats.st_mode))
	{
	  fatal_error ("'%s' exists but is not a directory", dirpath);
	  return (false);
	}
    }
  return (true);
}

_MIKTEX_1 void _MIKTEX_2
parse_web_filename (char *web_filename,
		    char **section,
		    char **key,
		    char **filename)

{
  if (*web_filename == '[')
    {
      web_filename = strtok (web_filename + 1, "]");
      if (web_filename)
	{
	  if (section)
	    *section = web_filename;
	  web_filename = strtok (0, ":");
	  if (web_filename)
	    {
	      if (*key)
		*key = web_filename;
	      if (filename)
		*filename = web_filename + strlen (web_filename) + 1;
	    }
	}
    }
  else if (filename)
    *filename = web_filename;
}

#define STRDUP(s) strcpy (_alloca (strlen (s) + 1), (s))

#define MAKETEXTFM_EXE "MakeTeXTFM.exe"

_MIKTEX_1 boolean _MIKTEX_2
mtopentfmfil (bytefile *f,
	      char *name)

{
  FILE *fp;
  char filenamebuf[_MAX_PATH];
  static char buf[2048];
  const char *path_list;
  char *section = "TeX";
  char *key = "Font Metric Dirs";
  char *filename = "fixme.tfm";
  char *save_name = STRDUP (name);
  parse_web_filename (save_name, &section, &key, &filename);
  path_list = get_cfg_value (section, key, buf, sizeof (buf), 0);
  if (path_list == 0)
    return (false);
  if (find_file (filename, path_list, filenamebuf))
    {
      fp = open_file (filenamebuf, c4prbmode);
      if (fp == 0)
	return (false);
      f->_c4p_fp = fp;
      get (*f);
      return (true);
    }
  else
    {
      char basename[_MAX_PATH];
      char command_line[_MAX_PATH];
      char temp_dir[_MAX_PATH];
      char root_dir[_MAX_PATH];
      char *argv[10];
      int argc = 0;
      char *ext;
      int rc;
      if (get_cfg_value ("TeX", "Font Metric Temp Dir",
			 temp_dir, sizeof (temp_dir), 0) == 0
	  || get_cfg_value ("MiKTeX", "Root Directory",
			    root_dir, sizeof (root_dir), 0) == 0)
	return (false);
      strcpy (basename, filename);
      ext = strstr (basename, ".tfm");
      if (ext)
	*ext = 0;
      strcpy (command_line, root_dir);
      strcat (command_line, "\\miktex\\bin\\");
      strcat (command_line, MAKETEXTFM_EXE);
      argv[argc++] = command_line;
      argv[argc++] = "-v";
      argv[argc++] = "-D";
      argv[argc++] = temp_dir;
      argv[argc++] = basename;
      argv[argc++] = 0;
      rc = _spawnv (_P_WAIT, command_line, argv);
      if (rc == 0 && find_file (filename, path_list, filenamebuf))
	{
	  fp = open_file (filenamebuf, c4prbmode);
	  if (fp == 0)
	    return (false);
	  f->_c4p_fp = fp;
	  get (*f);
	  return (true);
	}
    }
  return (false);
}

_MIKTEX_1 boolean _MIKTEX_2
mtopenvffile (bytefile *f,
	      char *name)

{
  FILE *fp;
  char filenamebuf[_MAX_PATH];
  static char buf[2048];
  const char *path_list;
  char *section = "dvips";
  char *key = "VFFONTS";
  char *filename = "fixme.vf";
  char *save_name = STRDUP (name);
  parse_web_filename (save_name, &section, &key, &filename);
  path_list = get_cfg_value (section, key, buf, sizeof (buf), 0);
  if (path_list == 0)
    return (false);
  if (find_file (filename, path_list, filenamebuf))
    {
      fp = open_file (filenamebuf, c4prbmode);
      if (fp == 0)
	return (false);
      f->_c4p_fp = fp;
      get (*f);
      return (true);
    }
  else
    return (false);
}

_MIKTEX_1 boolean _MIKTEX_2
find_pk_file (const char *fontname,
	      const char *mode,
	      int dpi,
	      char *result)

{
  char *search_path;
  char *p;
  const char *q;
  char basefilename[_MAX_PATH];
  char buf[MAX_PATH];
  strcpy (basefilename, fontname);
  strcat (basefilename, ".pk");
  q = get_cfg_value ("yap", "PK Input Dirs", buf, sizeof (buf), 0);
  if (q == 0)
    q = get_cfg_value ("dvips", "texpks", buf, sizeof (buf), 0);
  if (q == 0)
    q = "fixme";
  search_path = (char *) _alloca (strlen (q) + _MAX_PATH);
  p = search_path;
  while (*q)
    {
      if (*q == '%')
	{
	  q++;
	  if (*q == 0)
	    break;
	  switch (*q)
	    {
	    case '%':
	      *p++ = '%';
	      break;
	    case 'm':
	      if (mode)
		strcpy (p, mode);
	      else
		strcpy (p, "whatmode");
	      p += strlen (p);
	      break;
	    case 'd':
	      sprintf (p, "%ld", (long) dpi);
	      p += strlen (p);
	      break;
	    default:
	      strcpy (p, "whatcontrol");
	      p += strlen (p);
	      break;
	    }
	}
      else
	*p++ = *q;
      q++;
    }
  *p = 0;
  return (find_file (basefilename, search_path, result));
}

_MIKTEX_1 void _MIKTEX_2
miktexbanner (_c4p_textfile_t *f)

{
  fprintf (f->_c4p_fp,
	   " (%s %s)",
	   VER_PRODUCTNAME_STR, VER_PRODUCTVERSION_STR);
}

_MIKTEX_1 int _MIKTEX_2
set_metafont_mode (const mf_mode *mode)

{
  const char *home = get_home_directory (0);
  char dvisrc[_MAX_PATH];
  sprintf (dvisrc, "%s\\.dvipsrc", home);
  if (_access (dvisrc, 0))
    {
      char buf[1000];
      FILE *fp1, *fp2;
      char *tname = _tempnam ("c:\\tmp", "mik");
      fp1 = fopen (dvisrc, "r");
      if (fp1 == 0)
	return (-1);
      fp2 = fopen (tname, "w");
      if (fp2 == 0)
	{
	  fclose (fp1);
	  return (-1);
	}
      while (fgets (buf, sizeof (buf), fp1))
	{
	  if (*buf == 'M')
	    fprintf (fp2, "M %s\n", mode->mnemonic);
	  else if (*buf == 'D')
	    fprintf (fp2, "D %d", mode->mnemonic);
	  else
	    fputs (buf, fp2);
	}
      fclose (fp1);
      fclose (fp2);
      if (_unlink (dvisrc) == 0)
	rename (tname, dvisrc);
    }
  else
    {
      FILE *fp = fopen (dvisrc, "w");
      fprintf (fp, "M %s\n", mode);
      fclose (fp);
    }
}

_MIKTEX_1 int _MIKTEX_2
get_metafont_modes (mf_mode *modes,
		    size_t array_size)

{
  FILE *fp;
  char filenamebuf[_MAX_PATH];
  static char buf[2048];
  const char *path_list;
  char *section = "METAFONT";
  char *key = "Input Dirs";
  char *filename = "modes.mf";
  int count = 0;
  int in_mode_def = 0;
  path_list = get_cfg_value (section, key, buf, sizeof (buf), 0);
  if (path_list == 0)
    return (-1);
  if (! find_file (filename, path_list, filenamebuf))
    return (-1);
  fp = open_file (filenamebuf, "r");
  if (fp == 0)
    return (-1);
  while (fgets (buf, sizeof (buf), fp) && count < array_size)
    {
      if (in_mode_def)
	{
	  char *cp;
	  if (strncmp (buf, "enddef;", 7) == 0)
	    {
	      count++;
	      in_mode_def = 0;
	    }
	  else if (modes[count].hor_res == 0
		   && (cp = strstr (buf, "pixels_per_inch")))
	    {
	      while (*cp && ! isdigit (*cp))
		cp++;
	      modes[count].hor_res = atoi (cp);
	      modes[count].vert_res = atoi (cp);
	    }
	  else if (cp = strstr (buf, "aspect_ratio"))
	    {
	      while (*cp && ! isdigit (*cp))
		cp++;
	      modes[count].vert_res = atoi (cp);
	    }
	}
      else if (strncmp (buf, "mode_def", 8) == 0)
	{
	  char *cp = buf + 8;
	  char *mode_name;
	  char *printer_name;
	  while (! isalpha (*cp) && *cp != '\n')
	    cp++;
	  if (! isalpha (*cp))
	    continue;
	  mode_name = cp;
	  while (isalpha (*cp))
	    cp++;
	  if (*cp == 0)
	    continue;
	  *cp++ = 0;
	  if (strcmp (mode_name, "help") == 0)
	    continue;
	  cp = strstr (cp, "%\\[");
	  if (cp == 0)
	    continue;
	  cp += 3;
	  while (isspace (*cp))
	    *cp++;
	  if (cp == 0)
	    continue;
	  printer_name = cp;
	  while (*cp && *cp != '\n')
	    cp++;
	  *cp = 0;
	  in_mode_def++;
	  modes[count].mnemonic = strdup (mode_name);
	  modes[count].description = strdup (printer_name);
	  modes[count].hor_res = 0;
	}
    }
  return (count);
}

#ifdef MIKTEX_DLL
int APIENTRY 
DllMain (HINSTANCE hInstance,
	 DWORD dwReason,
	 LPVOID ignore)

{
  if (dwReason == DLL_PROCESS_ATTACH)
    {
      char buf[10];
      const char *cp = get_cfg_value ("miktex", "debug", buf, sizeof (buf), 0);
      if (cp && _stricmp (cp, "true") == 0)
	debug_flag = 1;
      else
	debug_flag = 0;
    }
  return (1);
}
#endif
