/* makempx.c - Create MPX files
   Time-stamp: "97/01/21 22:27:08 mik"

   Copyright (C) 1997
	Christian Schenk  <cschenk@berlin.snafu.de>

   This file is part of MiKTeX.

   MiKTeX is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   MiKTeX is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with MiKTeX; if not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <io.h>
#include <direct.h>
#include <process.h>
#include "sys/stat.h"

#include "miktex.h"
#include "makempx.rc"

#if defined (_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif /* _WIN32 */

static int verbose_p;
static int debug_p;
static int quiet_p;
static int print_only_p;
static const char *texdir;
static const char *destdir;
static char *name;

static char defdestdir[_MAX_PATH];
static char command_line[_MAX_PATH];

static char cwd[_MAX_PATH];
static char *temp_dir;

#if defined (_WIN32)
static char system_temp_dir[_MAX_PATH];
static int drive;
static int temp_drive;
#endif /* _WIN32 */

static struct option long_options[] =

{
  "debug",		no_argument,		0,	'd',
  "help",		no_argument,		0,	'h',
  "print-only",		no_argument,		0,	'n',
  "tex-dir",		required_argument,	0,	'T',
  "dest-dir",		required_argument,	0,	'D',
  "quiet",		no_argument,		0,	'q',
  "verbose",		no_argument,		0,	'v',
  "version",		no_argument,		0,	'V',
  0,			no_argument,		0,	0,
};

static void
show_version (void)

{
  printf ("This is makempx version %s (%s %s)\n",
	  VER_FILEVERSION_STR, VER_PRODUCTNAME_STR,
	  VER_PRODUCTVERSION_STR);
}

static void
usage (void)

{
  puts ("Usage: makempx MPFILE MPXFILE\n");

  puts ("This program makes a MPX file from a MP file.\n");
}

static void
clean ()

{
#if defined (_WIN32)
  if (print_only_p)
    printf ("cd %s\n", cwd);
  else
    {
      WIN32_FIND_DATA fdata;
      char path[_MAX_PATH];
      HANDLE hnd;
      sprintf (path, "%s\\*", temp_dir);
      hnd = FindFirstFile (path, &fdata);
      if (hnd != INVALID_HANDLE_VALUE)
	{
	  int proceed = 1;
	  while (proceed)
	    {
	      if (! (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
		{
		  sprintf (path, "%s\\%s", temp_dir, fdata.cFileName);
		  if (remove (path) != 0)
		    perror (path);
		}
	      proceed = FindNextFile (hnd, &fdata);
	    }
	  FindClose (hnd);
	}
      if (_chdrive (drive) != 0)
	perror (cwd);
      if (chdir (cwd) != 0)
	perror (cwd);
      if (rmdir (temp_dir) != 0)
	perror (temp_dir);
    }
#endif /* _WIN32 */
}

#define STRDUP(s) strcpy (_alloca (strlen (s) + 1), s)

static int
newer (const char *file1,
       const char *file2)

{
  struct stat x, y;
  if (stat (file1, &x) < 0)
    return (1);
  else if (stat (file2, &y) < 0)
    return (0);
  else if (x.st_mtime < y.st_mtime)
    return (1);
  else
    return (0);
}

int
main (int argc,
      char ** argv)

{
  int c;
  int option_index;
  char *tmpptr;
  int rc;
  char destname[_MAX_PATH];
  char *cp, *cp1;
  static char buf[2048];
  char *pargv[20];
  int pargc;
  char *mpname;
  char *mpxname;

  print_only_p = 0;
  quiet_p = 0;
#if defined (DEBUG)
  verbose_p = 1;
  debug_p = 1;
#else
  verbose_p = 0;
  debug_p = 0;
#endif
  texdir = 0;
  destdir = 0;

  c4pargv[0] = argv[0];
  while ((c=getopt_long (argc, argv, "dD:hnvV", long_options, &option_index))
	 != EOF)
    {
      switch (c)
	{
	case 'h':
	  usage ();
	  exit (0);
	  break;
	case 'n':
	  print_only_p = 1;
	  break;
	case 'T':
	  texdir = optarg;
	  break;
	case 'D':
	  destdir = optarg;
	  break;
	case 'd':
	  debug_p = 1;
	  break;
	case 'v':
	  verbose_p = 1;
	  break;
	case 'q':
	  quiet_p = 1;
	  break;
	case 'V':
	  show_version ();
	  exit (0);
	  break;
	default:
	  usage ();
	  exit (1);
	  break;
	}
    }

  if (argc - optind != 2)
    {
      usage ();
      exit (1);
    }

  mpname = argv[optind++];
  mpxname = argv[optind++];

  if (newer (mpname, mpxname))
    {
      exit (0);
    }

  if (texdir == 0)
    {
      char buf[MAX_PATH];
      texdir = get_cfg_value ("MiKTeX", "Root Directory",
			      buf, sizeof (buf), 0);
      if (texdir == 0)
	texdir = "c:\\texmf";
      else
	texdir = STRDUP (texdir);
    }

  if (getcwd (cwd, sizeof (cwd)) == 0)
    {
      perror ("getcwd");
      exit (1);
    }

#if defined (_WIN32)
  drive = _getdrive ();
  GetTempPath (sizeof (system_temp_dir), system_temp_dir);
  temp_dir = _tempnam (system_temp_dir, "mpx");
  if (temp_dir == 0)
    {
      perror ("_tempnam");
      exit (1);
    }
  temp_drive = tolower (temp_dir[0]) - 'a' + 1;
#endif /* _WIN32 */

#if 0
  if (destdir == 0)
    {
      char buf[MAX_PATH];
      destdir = get_cfg_value ("MetaPost", "MPX Temp Dir",
			       buf, sizeof (buf), 0);
      if (destdir == 0)
	destdir = ".";
      else
	destdir = STRDUP (destdir);
    }
#endif

  if (print_only_p)
    printf ("mkdir %s\n", temp_dir);
  else if (mkdir (temp_dir) != 0)
    {
      perror (temp_dir);
      exit (1);
    }
  
  if (verbose_p)
    printf ("Creating %s...\n", mpxname);

#if 0
  sprintf (command_line,
	   "%s\\miktex\\bin\\mptotex.exe %s >mpx314.tex 2>mpxerr.log",
	   texdir, mpname);
#else
  sprintf (command_line,
	   "%s\\miktex\\bin\\mptotex.exe %s >%s\\mpx314.tex",
	   texdir, mpname, temp_dir);
#endif

  if (! quiet_p || print_only_p)
    puts (command_line);

  if (print_only_p)
    rc = 0;
  else
    {
#if defined (_WIN32)
      _flushall ();
#endif
      rc = system (command_line);
    }
  if (rc < 0)
    {
      perror (command_line);
      clean ();
      exit (1);
    }
  if (rc != 0)
    {
      fprintf (stderr, "Command failed: %s", command_line);
      clean ();
      exit (1);
    }

#if defined (_WIN32)
  if (print_only_p)
    ;
  else if (_chdrive (temp_drive) != 0)
    {
      perror (temp_dir);
      clean ();
      exit (1);
    }
#endif /* _WIN32 */

  if (print_only_p)
    printf ("cd %s\n", temp_dir);
  else if (chdir (temp_dir) != 0)
    {
      perror (temp_dir);
      clean ();
      exit (1);
    }

#if 0
    if test -r ${MPTEXPRE:-mptexpre.tex}
    then
	cat ${MPTEXPRE:-mptexpre.tex} mpx$$.tex >mpx$$.tmp
	mv mpx$$.tmp mpx$$.tex
    fi
#endif

  sprintf (command_line,
	   "%s\\miktex\\bin\\tex.exe \\batchmode\\input mpx314>nul",
	   texdir);

  if (! quiet_p || print_only_p)
    puts (command_line);

  if (print_only_p)
    rc = 0;
  else
    {
#if defined (_WIN32)
      _flushall ();
#endif
      rc = system (command_line);
    }
  if (rc < 0)
    {
      perror (command_line);
      clean ();
      exit (1);
    }
  if (rc != 0)
    {
      fprintf (stderr, "Command failed: %s", command_line);
      clean ();
      exit (1);
    }

  if (_chdrive (drive) != 0)
    perror (cwd);

  if (chdir (cwd) != 0)
    perror (cwd);

  sprintf (command_line,
	   "%s\\miktex\\bin\\dvitomp.exe %s\\mpx314.dvi %s",
	   texdir, temp_dir, mpxname);

  if (! quiet_p || print_only_p)
    puts (command_line);

  if (print_only_p)
    rc = 0;
  else
    {
#if defined (_WIN32)
      _flushall ();
#endif
      rc = system (command_line);
    }
  if (rc < 0)
    {
      perror (command_line);
      clean ();
      exit (1);
    }
  if (rc != 0)
    {
      fprintf (stderr, "Command failed: %s", command_line);
      clean ();
      exit (1);
    }

  clean ();
  return (0);
}
